@@ -6,10 +6,36 @@
 //!
 //! Once parsed, the file can be traversed or modified, and then written back to
 //! a file.
+//!
+//! # Example
+//!
+//! ```rust
+//! use deb822_lossless::Deb822;
+//! use std::str::FromStr;
+//!
+//! let input = r#"Package: deb822-lossless
+//! Maintainer: Jelmer Vernooĳ <jelmer@debian.org>
+//! Homepage: https://github.com/jelmer/deb822-lossless
+//! Section: rust
+//!
+//! Package: deb822-lossless
+//! Architecture: any
+//! Description: Lossless parser for deb822 style files.
+//!   This parser can be used to parse files in the deb822 format, while preserving
+//!   all whitespace and comments. It is based on the [rowan] library, which is a
+//!   lossless parser library for Rust.
+//! "#;
+//!
+//! let deb822 = Deb822::from_str(input).unwrap();
+//! assert_eq!(deb822.paragraphs().count(), 2);
+//! let homepage = deb822.paragraphs().nth(0).unwrap().get("Homepage");
+//! assert_eq!(homepage.as_deref(), Some("https://github.com/jelmer/deb822-lossless"));
+//! ```
 
 mod lex;
 use crate::lex::lex;
 use rowan::ast::AstNode;
+use std::path::Path;
 use std::str::FromStr;
 
 /// Let's start with defining all kinds of tokens and
@@ -28,9 +54,10 @@ pub enum SyntaxKind {
     ERROR,      // as well as errors
 
     // composite nodes
-    ROOT,      // The entire file
-    PARAGRAPH, // A deb822 paragraph
-    ENTRY,     // A single key-value pair
+    ROOT,       // The entire file
+    PARAGRAPH,  // A deb822 paragraph
+    ENTRY,      // A single key-value pair
+    EMPTY_LINE, // An empty line
 }
 
 use SyntaxKind::*;
@@ -56,6 +83,35 @@ impl std::fmt::Display for ParseError {
 
 impl std::error::Error for ParseError {}
 
+#[derive(Debug)]
+pub enum Error {
+    ParseError(ParseError),
+    IoError(std::io::Error),
+}
+
+impl std::fmt::Display for Error {
+    fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
+        match &self {
+            Error::ParseError(err) => write!(f, "{}", err),
+            Error::IoError(err) => write!(f, "{}", err),
+        }
+    }
+}
+
+impl From<ParseError> for Error {
+    fn from(err: ParseError) -> Self {
+        Self::ParseError(err)
+    }
+}
+
+impl From<std::io::Error> for Error {
+    fn from(err: std::io::Error) -> Self {
+        Self::IoError(err)
+    }
+}
+
+impl std::error::Error for Error {}
+
 /// Second, implementing the `Language` trait teaches rowan to convert between
 /// these two SyntaxKind types, allowing for a nicer SyntaxNode API where
 /// "kinds" are values from our `enum SyntaxKind`, instead of plain u16 values.
@@ -117,7 +173,9 @@ fn parse(text: &str) -> Parse {
                 self.skip_ws();
             } else {
                 self.builder.start_node(ERROR.into());
-                self.bump();
+                if self.current().is_some() {
+                    self.bump();
+                }
                 self.errors.push("expected ':'".to_string());
                 self.builder.finish_node();
             }
@@ -163,7 +221,9 @@ fn parse(text: &str) -> Parse {
             self.builder.start_node(ROOT.into());
             while self.current().is_some() {
                 self.skip_ws_and_newlines();
-                self.parse_paragraph();
+                if self.current().is_some() {
+                    self.parse_paragraph();
+                }
             }
             // Don't forget to eat *trailing* whitespace
             self.skip_ws_and_newlines();
@@ -195,7 +255,14 @@ fn parse(text: &str) -> Parse {
                 || self.current() == Some(COMMENT)
                 || self.current() == Some(NEWLINE)
             {
-                self.bump()
+                self.builder.start_node(EMPTY_LINE.into());
+                while self.current() != Some(NEWLINE) && self.current().is_some() {
+                    self.bump();
+                }
+                if self.current() == Some(NEWLINE) {
+                    self.bump();
+                }
+                self.builder.finish_node();
             }
         }
     }
@@ -272,6 +339,12 @@ macro_rules! ast_node {
     };
 }
 
+impl std::fmt::Debug for Deb822 {
+    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
+        f.debug_struct("Deb822").finish()
+    }
+}
+
 ast_node!(Deb822, ROOT);
 ast_node!(Paragraph, PARAGRAPH);
 ast_node!(Entry, ENTRY);
@@ -296,21 +369,45 @@ impl Deb822 {
         self.0.children().filter_map(Paragraph::cast)
     }
 
+    /// Add a new empty paragraph to the end of the file.
     pub fn add_paragraph(&mut self) -> Paragraph {
         let paragraph = Paragraph::new();
         let mut to_insert = vec![];
         if self.0.children().count() > 0 {
             let mut builder = GreenNodeBuilder::new();
+            builder.start_node(EMPTY_LINE.into());
             builder.token(NEWLINE.into(), "\n");
-            to_insert.push(SyntaxNode::new_root(builder.finish()).clone_for_update().into());
+            builder.finish_node();
+            to_insert.push(
+                SyntaxNode::new_root(builder.finish())
+                    .clone_for_update()
+                    .into(),
+            );
         }
         to_insert.push(paragraph.0.clone().into());
         self.0.splice_children(
             self.0.children().count()..self.0.children().count(),
-            to_insert
+            to_insert,
         );
         paragraph
     }
+
+    pub fn from_file(path: impl AsRef<Path>) -> Result<Self, Error> {
+        let text = std::fs::read_to_string(path)?;
+        Ok(Self::from_str(&text)?)
+    }
+
+    pub fn from_file_relaxed(
+        path: impl AsRef<Path>,
+    ) -> Result<(Self, Vec<String>), std::io::Error> {
+        let text = std::fs::read_to_string(path)?;
+        Ok(Self::from_str_relaxed(&text))
+    }
+
+    pub fn from_str_relaxed(s: &str) -> (Self, Vec<String>) {
+        let parsed = parse(s);
+        (parsed.root(), parsed.errors)
+    }
 }
 
 impl Paragraph {
@@ -329,6 +426,7 @@ impl Paragraph {
             .map(|e| e.value())
     }
 
+    /// Returns whether the paragraph contains the given key.
     pub fn contains_key(&self, key: &str) -> bool {
         self.get(key).is_some()
     }
@@ -338,23 +436,30 @@ impl Paragraph {
         self.0.children().filter_map(Entry::cast)
     }
 
+    /// Returns an iterator over all items in the paragraph.
     pub fn items(&self) -> impl Iterator<Item = (String, String)> + '_ {
         self.entries()
             .filter_map(|e| e.key().map(|k| (k, e.value())))
     }
 
+    /// Returns an iterator over all values for the given key in the paragraph.
     pub fn get_all<'a>(&'a self, key: &'a str) -> impl Iterator<Item = String> + '_ {
-        self.items().filter_map(move |(k, v)| if k.as_str() == key { Some(v) } else { None })
+        self.items()
+            .filter_map(move |(k, v)| if k.as_str() == key { Some(v) } else { None })
     }
 
+    #[deprecated(note = "use `contains_key` instead")]
+    /// Returns true if the paragraph contains the given key.
     pub fn contains(&self, key: &str) -> bool {
         self.get_all(key).any(|_| true)
     }
 
+    /// Returns an iterator over all keys in the paragraph.
     pub fn keys(&self) -> impl Iterator<Item = String> + '_ {
         self.entries().filter_map(|e| e.key())
     }
 
+    /// Remove the given field from the paragraph.
     pub fn remove(&mut self, key: &str) {
         for mut entry in self.entries() {
             if entry.key().as_deref() == Some(key) {
@@ -363,10 +468,16 @@ impl Paragraph {
         }
     }
 
+    /// Add a new field to the paragraph.
     pub fn insert(&mut self, key: &str, value: &str) {
-        for mut entry in self.entries() {
+        let new_entry = Entry::new(key, value);
+
+        for entry in self.entries() {
             if entry.key().as_deref() == Some(key) {
-                entry.set_value(value);
+                self.0.splice_children(
+                    entry.0.index()..entry.0.index() + 1,
+                    vec![new_entry.0.clone_for_update().into()],
+                );
                 return;
             }
         }
@@ -377,12 +488,21 @@ impl Paragraph {
         );
     }
 
-    pub fn rename(&mut self, old_key: &str, new_key: &str) {
-        for mut entry in self.entries() {
+    /// Rename the given field in the paragraph.
+    pub fn rename(&mut self, old_key: &str, new_key: &str) -> bool {
+        for entry in self.entries() {
             if entry.key().as_deref() == Some(old_key) {
-                entry.set_key(new_key);
+                self.0.splice_children(
+                    entry.0.index()..entry.0.index() + 1,
+                    vec![Entry::new(new_key, entry.value().as_str())
+                        .0
+                        .clone_for_update()
+                        .into()],
+                );
+                return true;
             }
         }
+        false
     }
 }
 
@@ -400,7 +520,9 @@ impl std::str::FromStr for Paragraph {
 
         let mut paragraphs = deb822.paragraphs();
 
-        paragraphs.next().ok_or_else(||ParseError(vec!["no paragraphs".to_string()]))
+        paragraphs
+            .next()
+            .ok_or_else(|| ParseError(vec!["no paragraphs".to_string()]))
     }
 }
 
@@ -412,8 +534,13 @@ impl Entry {
         builder.token(KEY.into(), key);
         builder.token(COLON.into(), ":");
         builder.token(WHITESPACE.into(), " ");
-        builder.token(VALUE.into(), value);
-        builder.token(NEWLINE.into(), "\n");
+        for (i, line) in value.split("\n").enumerate() {
+            if i > 0 {
+                builder.token(INDENT.into(), " ");
+            }
+            builder.token(VALUE.into(), line);
+            builder.token(NEWLINE.into(), "\n");
+        }
         builder.finish_node();
         Entry(SyntaxNode::new_root(builder.finish()))
     }
@@ -436,14 +563,6 @@ impl Entry {
             .join("\n")
     }
 
-    pub fn set_key(&mut self, _key: &str) {
-        todo!();
-    }
-
-    pub fn set_value(&mut self, _value: &str) {
-        todo!();
-    }
-
     pub fn detach(&mut self) {
         self.0.detach();
     }
@@ -505,10 +624,13 @@ Description: This is a description
       WHITESPACE@58..59 " "
       VALUE@59..62 "net"
       NEWLINE@62..63 "\n"
-  NEWLINE@63..64 "\n"
-  COMMENT@64..83 "# This is a comment"
-  NEWLINE@83..84 "\n"
-  NEWLINE@84..85 "\n"
+  EMPTY_LINE@63..64
+    NEWLINE@63..64 "\n"
+  EMPTY_LINE@64..84
+    COMMENT@64..83 "# This is a comment"
+    NEWLINE@83..84 "\n"
+  EMPTY_LINE@84..85
+    NEWLINE@84..85 "\n"
   PARAGRAPH@85..203
     ENTRY@85..98
       KEY@85..92 "Package"
@@ -592,6 +714,51 @@ Description: This is a description
     assert_eq!(node.text(), CONTROLV1);
 }
 
+#[test]
+fn test_with_trailing_whitespace() {
+    const CONTROLV1: &str = r#"Source: foo
+Maintainer: Foo Bar <foo@example.com>
+
+
+"#;
+    let parsed = parse(CONTROLV1);
+    let node = parsed.syntax();
+    assert_eq!(
+        format!("{:#?}", node),
+        r###"ROOT@0..52
+  PARAGRAPH@0..50
+    ENTRY@0..12
+      KEY@0..6 "Source"
+      COLON@6..7 ":"
+      WHITESPACE@7..8 " "
+      VALUE@8..11 "foo"
+      NEWLINE@11..12 "\n"
+    ENTRY@12..50
+      KEY@12..22 "Maintainer"
+      COLON@22..23 ":"
+      WHITESPACE@23..24 " "
+      VALUE@24..49 "Foo Bar <foo@example. ..."
+      NEWLINE@49..50 "\n"
+  EMPTY_LINE@50..51
+    NEWLINE@50..51 "\n"
+  EMPTY_LINE@51..52
+    NEWLINE@51..52 "\n"
+"###
+    );
+    assert_eq!(parsed.errors, Vec::<String>::new());
+
+    let root = parsed.root();
+    assert_eq!(root.paragraphs().count(), 1);
+    let source = root.paragraphs().next().unwrap();
+    assert_eq!(
+        source.items().collect::<Vec<_>>(),
+        vec![
+            ("Source".into(), "foo".into()),
+            ("Maintainer".into(), "Foo Bar <foo@example.com>".into()),
+        ]
+    );
+}
+
 #[cfg(test)]
 mod tests {
     #[test]
@@ -623,7 +790,41 @@ Description: This is a description
     }
 
     #[test]
-    fn test_modify() {
+    fn test_after_multi_line() {
+        let d: super::Deb822 = r#"Source: golang-github-blah-blah
+Section: devel
+Priority: optional
+Standards-Version: 4.2.0
+Maintainer: Some Maintainer <example@example.com>
+Build-Depends: debhelper (>= 11~),
+               dh-golang,
+               golang-any
+Homepage: https://github.com/j-keck/arping
+"#
+        .parse()
+        .unwrap();
+        let mut ps = d.paragraphs();
+        let p = ps.next().unwrap();
+        assert_eq!(p.get("Source").as_deref(), Some("golang-github-blah-blah"));
+        assert_eq!(p.get("Section").as_deref(), Some("devel"));
+        assert_eq!(p.get("Priority").as_deref(), Some("optional"));
+        assert_eq!(p.get("Standards-Version").as_deref(), Some("4.2.0"));
+        assert_eq!(
+            p.get("Maintainer").as_deref(),
+            Some("Some Maintainer <example@example.com>")
+        );
+        assert_eq!(
+            p.get("Build-Depends").as_deref(),
+            Some("debhelper (>= 11~),\ndh-golang,\ngolang-any")
+        );
+        assert_eq!(
+            p.get("Homepage").as_deref(),
+            Some("https://github.com/j-keck/arping")
+        );
+    }
+
+    #[test]
+    fn test_remove_field() {
         let d: super::Deb822 = r#"Source: foo
 Maintainer: Foo Bar <jelmer@jelmer.uk>
 Section: net
@@ -650,4 +851,132 @@ Foo: Bar
 "#
         );
     }
+
+    #[test]
+    fn test_rename_field() {
+        let d: super::Deb822 = r#"Source: foo
+Vcs-Browser: https://salsa.debian.org/debian/foo
+"#
+        .parse()
+        .unwrap();
+        let mut ps = d.paragraphs();
+        let mut p = ps.next().unwrap();
+        assert!(p.rename("Vcs-Browser", "Homepage"));
+        assert_eq!(
+            p.to_string(),
+            r#"Source: foo
+Homepage: https://salsa.debian.org/debian/foo
+"#
+        );
+
+        assert_eq!(
+            p.get("Homepage").as_deref(),
+            Some("https://salsa.debian.org/debian/foo")
+        );
+        assert_eq!(p.get("Vcs-Browser").as_deref(), None);
+    }
+
+    #[test]
+    fn test_set_field() {
+        let d: super::Deb822 = r#"Source: foo
+Maintainer: Foo Bar <joe@example.com>
+"#
+        .parse()
+        .unwrap();
+        let mut ps = d.paragraphs();
+        let mut p = ps.next().unwrap();
+        p.insert("Maintainer", "Somebody Else <jane@example.com>");
+        assert_eq!(
+            p.get("Maintainer").as_deref(),
+            Some("Somebody Else <jane@example.com>")
+        );
+        assert_eq!(
+            p.to_string(),
+            r#"Source: foo
+Maintainer: Somebody Else <jane@example.com>
+"#
+        );
+    }
+
+    #[test]
+    fn test_set_new_field() {
+        let d: super::Deb822 = r#"Source: foo
+"#
+        .parse()
+        .unwrap();
+        let mut ps = d.paragraphs();
+        let mut p = ps.next().unwrap();
+        p.insert("Maintainer", "Somebody <joe@example.com>");
+        assert_eq!(
+            p.get("Maintainer").as_deref(),
+            Some("Somebody <joe@example.com>")
+        );
+        assert_eq!(
+            p.to_string(),
+            r#"Source: foo
+Maintainer: Somebody <joe@example.com>
+"#
+        );
+    }
+
+    #[test]
+    fn test_add_paragraph() {
+        let mut d = super::Deb822::new();
+        let mut p = d.add_paragraph();
+        p.insert("Foo", "Bar");
+        assert_eq!(p.get("Foo").as_deref(), Some("Bar"));
+        assert_eq!(
+            p.to_string(),
+            r#"Foo: Bar
+"#
+        );
+        assert_eq!(
+            d.to_string(),
+            r#"Foo: Bar
+"#
+        );
+
+        let mut p = d.add_paragraph();
+        p.insert("Foo", "Blah");
+        assert_eq!(p.get("Foo").as_deref(), Some("Blah"));
+        assert_eq!(
+            d.to_string(),
+            r#"Foo: Bar
+
+Foo: Blah
+"#
+        );
+    }
+
+    #[test]
+    fn test_multiline_entry() {
+        use super::SyntaxKind::*;
+        use rowan::ast::AstNode;
+
+        let entry = super::Entry::new("foo", "bar\nbaz");
+        let tokens: Vec<_> = entry.syntax()
+                                  .descendants_with_tokens()
+                                  .filter_map(|tok| tok.into_token())
+                                  .collect();
+
+        assert_eq!("foo: bar\n baz\n", entry.to_string());
+        assert_eq!("bar\nbaz", entry.value());
+
+        assert_eq!(
+            vec![
+                (KEY, "foo"),
+                (COLON, ":"),
+                (WHITESPACE, " "),
+                (VALUE, "bar"),
+                (NEWLINE, "\n"),
+                (INDENT, " "),
+                (VALUE, "baz"),
+                (NEWLINE, "\n"),
+            ],
+            tokens
+                .iter()
+                .map(|token| (token.kind(), token.text()))
+                .collect::<Vec<_>>()
+        );
+    }
 }
